from tabulate import tabulate
from texttable import Texttable
from pandas import DataFrame
from .openbis_object import OpenBisObject
from .things import Things
from .utils import check_datatype, split_identifier, format_timestamp, is_identifier, is_permid, nvl, extract_permid, extract_code, extract_name, VERBOSE
from .definitions import get_method_for_entity, get_type_for_entity, get_definition_for_entity
from .semantic_annotation import SemanticAnnotation

class EntityType:
    """ EntityTypes define a variety of an entity, eg. sample, dataSet, experiment
    This is the parent class of the SampleType, DataSetType, ExperimentType and
    MaterialType classes.
    """ 

    def __init__(self, openbis_obj, data=None, method=None, **kwargs):
        """This __init__ is called by OpenBisObject.__init__
        It stores the propertyAssignments data into the _propertyAssignments
        dict
        """
        self._set_entity_data(data)
        if method:
            self.__dict__['_get_method'] = method
            
    def _set_entity_data(self, data=None):
        pas = []
        if data is not None and 'propertyAssignments' in data: 
            pas = data['propertyAssignments'] 
        self.__dict__['_propertyAssignments'] = pas

    def __str__(self):

        """String representation of this entity type
        """
        return self.data['code']

    def _attrs(self):
        return [
            'code', 'description', 'autoGeneratedCode', 'subcodeUnique',
            'generatedCodePrefix', 'listable', 'showContainer', 'showParents',
            'showParentMetadata', 'validationPlugin',
        ]

    def __dir__(self):
        defs = get_definition_for_entity(self.entity)
        attrs = [
            'get_property_assignments()',
            'assign_property()',
            'revoke_property()',
            'move_property_to_top()',
            'move_property_after()',
            'get_validationPlugin()',
            'save()',
            'delete()',
        ]
        if self.is_new:
            return attrs + defs['attrs_new']
        else:
            return attrs + list(set(defs['attrs'] + defs['attrs_up']))

    def __getattr__(self, name):
        if name in self._attrs():
            if name in self.data:
                return self.data[name]
            else:
                return ''

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)

    def get_property_assignments(self):
        attrs = [
            'propertyType',
            'section', 'ordinal',
            'mandatory', 'initialValueForExistingEntities', 
            'showInEditView', 'showRawValueInForms',
            'registrator', 'registrationDate', 'plugin'
        ]
        
        pas = self.__dict__['_propertyAssignments']
        df = DataFrame(pas, columns=attrs)
        df['propertyType'] = df['propertyType'].map(extract_code)
        df['plugin'] = df['plugin'].map(extract_name)
        df['registrationDate'] = df['registrationDate'].map(format_timestamp)

        return Things(
            openbis_obj = self.openbis,
            entity = 'propertyType',
            single_item_method = self.openbis.get_property_type,
            identifier_name = 'propertyType',
            df = df,
            start_with = 1,
            count = len(pas),
            totalCount = len(pas),
        )

    def assign_property(self, 
        prop, plugin=None,
        section=None, ordinal=None,
        mandatory=False, initialValueForExistingEntities=None,
        showInEditView=True, showRawValueInForms=True

    ):
        """The «section» groups certain properties.
        The «ordinal» is defining the rank in the list where the property appears.
        The «mandatory» defines whether a property must be filled in. If you make a
        property mandatory after you already created entities, you have to define an
        «initialValueForExistingEntities» too.
        """
        if self.is_new:
            raise ValueError("Please save {} first".format(self.entity))

        if isinstance(prop, str):
            property_type = self.openbis.get_property_type(prop.upper())
        else:
            property_type = prop

        new_assignment = {
            "section": section,
            "ordinal": ordinal,
            "mandatory": mandatory,
            "initialValueForExistingEntities": initialValueForExistingEntities,
            "showInEditView": showInEditView,
            "showRawValueInForms": showRawValueInForms,
            "propertyTypeId": {
                "@type": "as.dto.property.id.PropertyTypePermId",
                "permId": property_type.permId,
            },
            "@type": "as.dto.property.create.PropertyAssignmentCreation",
        }

        # assign plugin
        if plugin is not None:
            plugin_obj = self.openbis.get_plugin(plugin)
            new_assignment['plugin'] = plugin_obj.name

        request = self._get_request_for_pa(new_assignment, 'Add')
        try:
            self.openbis._post_request(self.openbis.as_v3, request)
            new_data = self._get_method(self.permId, only_data=True) 
            self._set_entity_data(new_data)
            if VERBOSE:
                print(f"Property {property_type.permId} assigned to {self.permId}")
        except ValueError as exc:
            if 'already assigned' in str(exc):
                if VERBOSE: 
                    print(f"Property {property_type.permId} already assigned to {self.permId}")
            else:
                raise ValueError(exc)


    def revoke_property(self, prop, force=False):
        if isinstance(prop, str):
            property_type = prop.upper()
        else:
            property_type = prop.permId.upper()

        items = {
            "entityTypeId": self._permId,
            "propertyTypeId": {
                "permId": property_type,
                "@type" : "as.dto.property.id.PropertyTypePermId"
            },
            "@type": "as.dto.property.id.PropertyAssignmentPermId"
        }
        request = self._get_request_for_pa(items, 'Remove', force)
        resp  = self.openbis._post_request(self.openbis.as_v3, request)
        if not resp and VERBOSE:
            new_data = self._get_method(self.permId, only_data=True) 
            self._set_entity_data(new_data)

            print(f"Property {property_type} revoked from {self.permId}")
        

    def _get_request_for_pa(self, items, item_action, force=False):

        method  = get_method_for_entity(self.entity, 'update')
        entity_type = get_type_for_entity(self.entity, 'update')

        list_action_type = "as.dto.common.update.ListUpdateAction"+item_action

        request= {
            "method": method,
            "params": [
                self.openbis.token,
                [{
                    "propertyAssignments": {
                        "actions": [{
                            "items": [{
                                **items,
                            }],
                            "@type": list_action_type
                        }],
                        "@type": "as.dto.entitytype.update.PropertyAssignmentListUpdateValue",
                        "forceRemovingAssignments": force
                    },
                    **entity_type,
                    "typeId": self._permId
                }]
            ]
        }

        return request
        
    def move_property_to_top(self, property):
        raise ValueError("not implemented yet")

    def move_property_after(self, property, after_property):
        raise ValueError("not implemented yet")


    @property
    def get_validationPlugin(self):
        """Returns a validation plugin object when called.
        Returns None when no validation plugin is defined.
        """
        try:
            return self.openbis.get_plugin(self._validationPlugin['name'])
        except Exception:
            pass

    def codes(self):
        codes = []
        for pa in self.data['propertyAssignments']:
            codes.append(pa['propertyType']['code'].lower())
        return codes


class SampleType(
    OpenBisObject, EntityType,
    entity='sampleType',
    single_item_method_name='get_sample_type'
):

    def __init__(self, openbis_obj, type=None, data=None, props=None, method=None, **kwargs):
        OpenBisObject.__init__(self, openbis_obj, type=type, data=data, props=props, **kwargs)
        EntityType.__init__(self, openbis_obj, type=type, data=data, props=props, method=method, **kwargs)
        

    def __dir__(self):
        return [
            'add_semantic_annotation()',
            'get_semantic_annotations()'
        ] + EntityType.__dir__(self) + OpenBisObject.__dir__(self)


    def add_semantic_annotation(self, **kwargs):
        semantic_annotation = SemanticAnnotation(
            openbis_obj=self.openbis, isNew=True, 
            entityType=self.code, **kwargs
        )
        semantic_annotation.save()
        return semantic_annotation

    def get_semantic_annotations(self):
        return self.openbis.search_semantic_annotations(entityType=self.code)

class DataSetType(
    OpenBisObject, EntityType,
    entity='dataSetType',
    single_item_method_name='get_dataset_type'
):

    def __init__(self, openbis_obj, type=None, data=None, props=None, method=None, **kwargs):
        OpenBisObject.__init__(self, openbis_obj, type=type, data=data, props=props, **kwargs)
        EntityType.__init__(self, openbis_obj, type=type, data=data, props=props, method=method, **kwargs)

    def __dir__(self):
        return [] + EntityType.__dir__(self) + OpenBisObject.__dir__(self)


class MaterialType(
    OpenBisObject, EntityType,
    entity='materialType',
    single_item_method_name='get_material_type'
):
    def __init__(self, openbis_obj, type=None, data=None, props=None, method=None, **kwargs):
        OpenBisObject.__init__(self, openbis_obj, type=type, data=data, props=props, **kwargs)
        EntityType.__init__(self, openbis_obj, type=type, data=data, props=props, method=method, **kwargs)

    def __dir__(self):
        return [
        ] + EntityType.__dir__(self) + OpenBisObject.__dir__(self)


class ExperimentType(
    OpenBisObject, EntityType,
    entity='experimentType',
    single_item_method_name='get_experiment_type'
):

    def __init__(self, openbis_obj, type=None, data=None, props=None, method=None, **kwargs):
        OpenBisObject.__init__(self, openbis_obj, type=type, data=data, props=props, **kwargs)
        EntityType.__init__(self, openbis_obj, type=type, data=data, props=props, method=method, **kwargs)

    def __dir__(self):
        return [
        ] + EntityType.__dir__(self) + OpenBisObject.__dir__(self)



from .property import PropertyHolder
from .attribute import AttrHolder
from .openbis_object import OpenBisObject 
from .definitions import openbis_definitions
from .utils import VERBOSE

class Sample(
    OpenBisObject,
    entity='sample',
    single_item_method_name='get_sample'
):
    """ A Sample (new: Object) is one of the most commonly used entities in openBIS.
    """

    def __init__(self, openbis_obj, type, project=None, data=None, props=None, **kwargs):
        self.__dict__['openbis'] = openbis_obj
        self.__dict__['type'] = type
        ph = PropertyHolder(openbis_obj, type)
        self.__dict__['p'] = ph
        self.__dict__['props'] = ph
        self.__dict__['a'] = AttrHolder(openbis_obj, 'sample', type)

        if data is not None:
            self._set_data(data)

        if project is not None:
            setattr(self, 'project', project)

        if props is not None:
            for key in props:
                setattr(self.p, key, props[key])

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

            if 'experiment' in kwargs:
                try:
                    experiment = getattr(self, 'experiment')
                    if not 'space' in kwargs:
                        project = experiment.project
                        setattr(self.a, 'space', project.space)
                except Exception:
                    pass

        if getattr(self, 'parents') is None:
            self.a.__dict__['_parents'] = []
        else:
            if not self.is_new:
                self.a.__dict__['_parents_orig'] = self.a.__dict__['_parents']

        if getattr(self, 'children') is None:
            self.a.__dict__['_children'] = []
        else:
            if not self.is_new:
                self.a.__dict__['_children_orig'] = self.a.__dict__['_children']

        if getattr(self, 'components') is None:
            self.a.__dict__['_components'] = []
        else:
            if not self.is_new:
                self.a.__dict__['_components_orig'] = self.a.__dict__['_components']


    def _set_data(self, data):
        # assign the attribute data to self.a by calling it
        # (invoking the AttrHolder.__call__ function)
        self.a(data)
        self.__dict__['data'] = data

        # put the properties in the self.p namespace (without checking them)
        for key, value in data['properties'].items():
            self.p.__dict__[key.lower()] = value


    def __dir__(self):
        return [
            'type',
            'get_parents()', 'get_children()', 'get_components()',
            'add_parents()', 'add_children()', 'add_components()', 
            'del_parents()', 'del_children()', 'del_components()',
            'set_parents()', 'set_children()', 'set_components()',
            'get_datasets()', 
            'space', 'project', 'experiment', 'container', 'tags',
            'set_tags()', 'add_tags()', 'del_tags()',
            'add_attachment()', 'get_attachments()', 'download_attachments()',
            'save()', 'delete()', 'mark_to_be_deleted()', 'unmark_to_be_deleted()', 'is_marked_to_be_deleted()',
            'attrs',
            'props',
        ] + super().__dir__()


    def _container(self, value=None):
        if value is not None:
            if value == '':
                if self.is_new:
                    pass
                else:
                    self.a.__dict__['_container'] = {}
            else:
                obj = None
                if isinstance(value, str):
                    # fetch object in openBIS, make sure it actually exists
                    obj = getattr(self._openbis, "get_sample")(value)
                elif value is None:
                    self.a.__dict__['_container'] = {}
                else:
                    obj = value

                self.a.__dict__['_container'] = obj.data['identifier']

                # mark attribute as modified, if it's an existing entity
                if self.is_new:
                    pass
                else:
                    self.a.__dict__['_container']['isModified'] = True
        else:
            try:
                return self.openbis.get_sample(self.a._container['identifier'])
            except Exception:
                pass

    @property
    def type(self):
        return self.__dict__['type']

    @type.setter
    def type(self, type_name):
        sample_type = self.openbis.get_sample_type(type_name)
        self.p.__dict__['_type'] = sample_type
        self.a.__dict__['_type'] = sample_type

    def __getattr__(self, name):
        if name in ['container']:
            return getattr(self, "_"+name)()

        return getattr(self.__dict__['a'], name)

    def __setattr__(self, name, value):
        if name in ['set_properties', 'set_tags', 'add_tags']:
            raise ValueError("These are methods which should not be overwritten")

        if name in ['container']:
            return getattr(self, "_"+name)(value)

        if name in ['p', 'props']:
            if isinstance(value, dict):
                for p in value:
                    setattr(self.__dict__['p'], p, value[p])
            else:
                raise ValueError("please provide a dictionary for setting properties")
        else:
            # must be an attribute in the AttributeHolder class
            setattr(self.__dict__['a'], name, value)

    def _repr_html_(self):
        return self.a._repr_html_()

    def __repr__(self):
        return self.a.__repr__()

    def set_properties(self, properties):
        for prop in properties.keys():
            setattr(self.p, prop, properties[prop])

    set_props = set_properties

    def get_datasets(self, **kwargs):
        return self.openbis.get_datasets(sample=self.permId, **kwargs)

    def get_projects(self, **kwargs):
        return self.openbis.get_project(withSamples=[self.permId], **kwargs)

    @property
    def experiment(self):
        try:
            return self.openbis.get_experiment(self._experiment['identifier'])
        except Exception:
            pass





import sys

if sys.version_info < (3,6):
    sys.exit('Sorry, Python < 3.6 is not supported')

from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name='PyBIS',
    version= '1.19.0RC0',
    author='Swen Vermeul • ID SIS • ETH Zürich',
    author_email='swen@ethz.ch',
    description='openBIS connection and interaction, optimized for using with Jupyter',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://sissource.ethz.ch/sispub/openbis/tree/master/pybis',
    packages=find_packages(),
    license='Apache Software License Version 2.0',
    install_requires=[
        'pytest',
        'requests',
        'datetime',
        'pandas',
        'click',
        'texttable',
        'tabulate',
    ],
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
